package net.darkmist.unicode;

import java.io.File;
import java.io.IOException;

public class Utf8 extends Util	// I'm so lazy
{
	protected String prevDisp = null;
	protected String prevStr = null;
	protected byte[] prevBytes = null;
	protected int prevDups = 0;
	protected char[] prevChars = null;
	private static final int COL_BYTES	= 12;
	private static final int COL_STR	= 26;
	private static final int COL_ASCII	= 11;
	private static final int COL_UTF16	= 23;
	private static final int COL_UTF8	= 35;
	private static final String HEADER = prefix("Bytes:") + suffix("Decoded:","ASCII:","UTF-16:","UTF-8:","FS:");
	private static final String ELIPSES = prefix("...") + suffix("...","...","...","...","");

	private static String suffix(String str, String ascii, String utf16, String utf8, String fs)
	{
		StringBuffer sb = new StringBuffer();

		// col 1: max 12
		// col 2: max 14
		// col 3: max 12
		// col 4: max 2 but title is 6
		// col 5: max 12
		// col 6: max 18
		// f0 80 80 80  | \ufffd\ufffd\ufffd\ufffd | 3f 3f 3f 3f | ff fd ff fd ff fd ff fd | ef bf bd ef bf bd ef bf bd ef bf bd |
		sb.append(" | ");
		rpad(sb,str,COL_STR);
		sb.append(" | ");
		rpad(sb,ascii,COL_ASCII);
		sb.append(" | ");
		rpad(sb,utf16,COL_UTF16);
		sb.append(" | ");
		rpad(sb,utf8,COL_UTF8);
		sb.append(" | ");
		sb.append(fs);
		return sb.toString();
	}

	private StringBuffer dispSuffix(StringBuffer sb, String str)
	{
		int preLen;

		sb.append(" | ");

		preLen = sb.length();
		sb.append('"');
		dispStr(sb,str);
		sb.append('"');
		rpad(sb,preLen,COL_STR);

		sb.append(" | ");

		preLen = sb.length();
		hexStr(sb, str2AsciiBytes(str));
		rpad(sb,preLen,COL_ASCII);

		sb.append(" | ");

		preLen = sb.length();
		hexStr(sb, str2Utf16Bytes(str));
		rpad(sb,preLen,COL_UTF16);

		sb.append(" | ");

		preLen = sb.length();
		hexStr(sb, str2Utf8Bytes(str));
		rpad(sb,preLen,COL_UTF8);

		sb.append(" | ");

		getFSTest(sb,str);
		return sb;
	}

	private static String prefix(String s)
	{
		return rpad(s,COL_BYTES);
	}

	private static String prefix(byte[] bytes)
	{
		return rpad(hexStr(new StringBuffer(), bytes),0,COL_BYTES).toString();
	}

	private static StringBuffer getFSTest(StringBuffer sb, String s)
	{
		String filename = "/" + s;
		File file = new File(filename);

		if(!file.exists())
			return sb;
		if(file.isDirectory())
			sb.append("dir ");
		else
			sb.append("exists ");
		try
		{
			String canonical = file.getCanonicalPath();
			if(!filename.equals(canonical))
			{
				sb.append('"');
				dispStr(sb,filename);
				sb.append("\"!=\"");
				dispStr(sb,canonical);
				sb.append('"');
			}
		}
		catch(IOException e)
		{
			sb.append(e.toString());
		}
		return sb;
	}

	private void dispBytes2Utf8(byte[] bytes)
	{
		String s;

		s = bytes2utf8(bytes);

		if(prevStr!=null)
		{
			if(s.equals(prevStr))
			{	// duplicate, copy bytes for end of dups
				prevBytes = byteDup(bytes,prevBytes);
				prevDups++;
				return;
			}
			else if(prevDups>0)
			{	// non duplicate
				if(prevDups>1)
					System.out.println(ELIPSES);
				System.out.print(prefix(prevBytes));
				System.out.println(prevDisp);
			}
		}
		prevDups = 0;
		prevStr = s;
		System.out.print(prefix(bytes));
		prevDisp = dispSuffix(new StringBuffer(),s).toString();
		System.out.println(prevDisp);
	}

	private void one(int min, int max)
	{
		int c;
		byte[] bytes = new byte[1];

		for(c=min;c<=max;c++)
		{
			bytes[0] = (byte)(c);
			dispBytes2Utf8(bytes);
		}
	}

	private void one()
	{
		one(0,255);
	}

	private void two(int min, int max)
	{
		int c;
		byte[] bytes = new byte[2];

		bytes[0] = (byte)(0xc0);
		for(c=min;c<=max;c++)
		{
			bytes[0] = (byte)(((c>>6)&0x1f)|0xc0);
			bytes[1] = (byte)((c&0x3f)|0x80);
			dispBytes2Utf8(bytes);
		}
	}

	private void two()
	{
		two(0,0x7ff);
	}

	private void three(int min, int max)
	{
		int c;
		byte[] bytes = new byte[3];

		bytes[0] = (byte)(0xc0);
		for(c=min;c<max;c++)
		{
			bytes[0] = (byte)(((c>>12)&0x0f)|0xe0);
			bytes[1] = (byte)(((c>>6)&0x3f)|0x80);
			bytes[2] = (byte)((c&0x3f)|0x80);
			dispBytes2Utf8(bytes);
		}
	}

	private void three()
	{
		three(0,0x10000);
	}

	private void four(int min, int max)
	{
		int c;
		byte[] bytes = new byte[4];

		bytes[0] = (byte)(0xc0);
		for(c=min;c<max;c++)
		{
			bytes[0] = (byte)(((c>>18)&0x0f)|0xf0);
			bytes[1] = (byte)(((c>>12)&0x3f)|0x80);
			bytes[2] = (byte)(((c>>6)&0x3f)|0x80);
			bytes[3] = (byte)((c&0x3f)|0x80);
			dispBytes2Utf8(bytes);
		}
	}

	private void four()
	{
		four(0,0x10FFFF);
	}

	/**
	 * Instance main method.
	 * @param args Ignored arguments to main.
	 */
	public void nonStaticMain(String args[])
	{
		//one(0,255);
		//two(0,255);
		//three(0,255);
		System.out.println(HEADER);
		one();
		two();
		three();
		four();
	}

	/**
	 * Main Method. Instantiates and calls {@link #nonStaticMain(String[])}.
	 * @param args
	 */
	public static void main(String args[])
	{
		new Utf8().nonStaticMain(args);
	}
}
